from __future__ import annotations

import json
import logging
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING

import polars as pl
from tqdm import tqdm

if TYPE_CHECKING:
    from board_game_merger.config import MergeConfig

LOGGER = logging.getLogger(__name__)
MAX_DISPLAY_ITEMS = 10


def merge_files(
    *,
    merge_config: MergeConfig,
    overwrite: bool = False,
    drop_empty: bool = False,
    sort_keys: bool = False,
    progress_bar: bool = False,
) -> None:
    """
    Merge files into one. Execute the following steps:

    - Filter out rows older than latest_min
    - For each row with identical keys, keep the latest one
    - Sort the output by keys, latest, or fields
    - Select only specified fields or exclude some fields
    - For each row, remove empty fields and sort keys alphabetically
    """

    if (
        merge_config.fieldnames_include is not None
        and merge_config.fieldnames_exclude is not None
    ):
        msg = "Cannot specify both fieldnames_include and fieldnames_exclude"
        raise ValueError(msg)

    in_paths_iter = (
        merge_config.in_paths
        if isinstance(merge_config.in_paths, list)
        else [merge_config.in_paths]
    )
    in_paths = [Path(in_path).resolve() for in_path in in_paths_iter]
    out_path = Path(merge_config.out_path).resolve()

    LOGGER.info(
        "Merging items from %s into <%s>",
        f"[{len(in_paths)} paths]" if len(in_paths) > MAX_DISPLAY_ITEMS else in_paths,
        out_path,
    )

    if not overwrite and out_path.exists():
        LOGGER.warning("Output file already exists, use overwrite to replace it")
        return

    data = pl.scan_ndjson(
        source=in_paths,
        schema=merge_config.schema,
        batch_size=512,
        low_memory=True,
        rechunk=True,
        ignore_errors=True,
    )

    latest_col = (
        merge_config.latest_col
        if isinstance(merge_config.latest_col, list)
        else [merge_config.latest_col]
    )
    if merge_config.latest_min is not None:
        LOGGER.info("Filtering out rows before <%s>", merge_config.latest_min)
        data = data.filter(latest_col[0] >= merge_config.latest_min)

    key_col = (
        merge_config.key_col
        if isinstance(merge_config.key_col, list)
        else [merge_config.key_col]
    )
    key_col_dict = {f"__key__{i}": key for i, key in enumerate(key_col)}

    LOGGER.info("Merging rows with identical keys: %s", key_col)
    LOGGER.info("Keeping latest by: %s", latest_col)

    data = (
        data.sort(by=latest_col, descending=True, nulls_last=True)
        .with_columns(**key_col_dict)
        .unique(subset=list(key_col_dict), keep="first")
        .drop(key_col_dict.keys())
    )

    if merge_config.sort_fields is not None:
        LOGGER.info(
            "Sorting data by: %s (%s)",
            merge_config.sort_fields,
            "descending" if merge_config.sort_descending else "ascending",
        )
        data = data.sort(
            merge_config.sort_fields,
            descending=merge_config.sort_descending,
        )

    if merge_config.fieldnames_include is not None:
        LOGGER.info("Selecting fields: %s", merge_config.fieldnames_include)
        data = data.select(merge_config.fieldnames_include)
    elif merge_config.fieldnames_exclude is not None:
        LOGGER.info("Excluding fields: %s", merge_config.fieldnames_exclude)
        data = data.select(pl.exclude(merge_config.fieldnames_exclude))

    LOGGER.info("Collecting results, this may take a while…")
    result = data.collect()
    LOGGER.info("Finished collecting results with shape %dx%d", *result.shape)
    num_rows = len(result)

    if not drop_empty and not sort_keys:
        LOGGER.info("Writing merged data to <%s>", out_path)
        result.write_ndjson(out_path)
        LOGGER.info("Done.")
        return

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = Path(temp_dir) / "merged.jl"
        LOGGER.info("Writing merged data to <%s>", temp_file)
        result.write_ndjson(temp_file)
        del result

        LOGGER.info("Writing cleaned data to <%s>", out_path)
        with temp_file.open("r") as in_file, out_path.open("w") as out_file:
            lines = (
                tqdm(
                    in_file,
                    desc="Cleaning data",
                    unit=" rows",
                    total=num_rows,
                )
                if progress_bar
                else in_file
            )
            for line in lines:
                row = json.loads(line)
                if drop_empty:
                    row = {k: v for k, v in row.items() if v}
                json.dump(row, out_file, sort_keys=sort_keys, separators=(",", ":"))
                out_file.write("\n")

    LOGGER.info("Done.")



import polars as pl

GAME_ITEM_SCHEMA = pl.Schema(
    {  # type: ignore[arg-type]
        "name": pl.String,
        "alt_name": pl.List(pl.String),
        "year": pl.Int64,
        "game_type": pl.List(pl.String),
        "description": pl.String,
        "designer": pl.List(pl.String),
        "artist": pl.List(pl.String),
        "publisher": pl.List(pl.String),
        "url": pl.String,
        "official_url": pl.List(pl.String),
        "image_url": pl.List(pl.String),
        "image_url_download": pl.List(pl.String),
        "image_file": pl.List(
            pl.Struct(
                {
                    "url": pl.String,
                    "path": pl.String,
                    "checksum": pl.String,
                },
            ),
        ),
        "image_blurhash": pl.List(
            pl.Struct(
                {
                    "url": pl.String,
                    "path": pl.String,
                    "checksum": pl.String,
                    "blurhash": pl.String,
                },
            ),
        ),
        "video_url": pl.List(pl.String),
        "rules_url": pl.List(pl.String),
        "rules_file": pl.List(
            pl.Struct(
                {
                    "url": pl.String,
                    "path": pl.String,
                    "checksum": pl.String,
                },
            ),
        ),
        "review_url": pl.List(pl.String),
        "external_link": pl.List(pl.String),
        "list_price": pl.String,
        "min_players": pl.Int64,
        "max_players": pl.Int64,
        "min_players_rec": pl.Int64,
        "max_players_rec": pl.Int64,
        "min_players_best": pl.Int64,
        "max_players_best": pl.Int64,
        "min_age": pl.Int64,
        "max_age": pl.Int64,
        "min_age_rec": pl.Float64,
        "max_age_rec": pl.Float64,
        "min_time": pl.Int64,
        "max_time": pl.Int64,
        "category": pl.List(pl.String),
        "mechanic": pl.List(pl.String),
        "cooperative": pl.Boolean,
        "compilation": pl.Boolean,
        # "compilation_of": pl.List(pl.Int64),
        "family": pl.List(pl.String),
        "expansion": pl.List(pl.String),
        "implementation": pl.List(pl.Int64),
        "integration": pl.List(pl.Int64),
        "rank": pl.Int64,
        "add_rank": pl.List(
            pl.Struct(
                {
                    "game_type": pl.String,
                    "game_type_id": pl.Int64,
                    "name": pl.String,
                    "rank": pl.Int64,
                    "bayes_rating": pl.Float64,
                },
            ),
        ),
        "num_votes": pl.Int64,
        "avg_rating": pl.Float64,
        "stddev_rating": pl.Float64,
        "bayes_rating": pl.Float64,
        "worst_rating": pl.Int64,
        "best_rating": pl.Int64,
        "complexity": pl.Float64,
        "easiest_complexity": pl.Int64,
        "hardest_complexity": pl.Int64,
        "language_dependency": pl.Float64,
        "lowest_language_dependency": pl.Int64,
        "highest_language_dependency": pl.Int64,
        "bgg_id": pl.Int64,
        "freebase_id": pl.String,
        "wikidata_id": pl.String,
        "wikipedia_id": pl.String,
        "dbpedia_id": pl.String,
        "luding_id": pl.Int64,
        "spielen_id": pl.String,
        "published_at": pl.String,
        "updated_at": pl.String,
        "scraped_at": pl.String,
    },
)


USER_ITEM_SCHEMA = pl.Schema(
    {  # type: ignore[arg-type]
        "item_id": pl.Int64,
        "bgg_user_name": pl.String,
        "first_name": pl.String,
        "last_name": pl.String,
        "registered": pl.Int64,
        "last_login": pl.String,
        "country": pl.String,
        "region": pl.String,
        "city": pl.String,
        "external_link": pl.List(pl.String),
        "image_url": pl.List(pl.String),
        "image_url_download": pl.List(pl.String),
        "image_file": pl.List(
            pl.Struct(
                {
                    "url": pl.String,
                    "path": pl.String,
                    "checksum": pl.String,
                },
            ),
        ),
        "image_blurhash": pl.List(
            pl.Struct(
                {
                    "url": pl.String,
                    "path": pl.String,
                    "checksum": pl.String,
                    "blurhash": pl.String,
                },
            ),
        ),
        "published_at": pl.String,
        "updated_at": pl.String,
        "scraped_at": pl.String,
    },
)

RATING_ITEM_SCHEMA = pl.Schema(
    {
        "item_id": pl.String,
        "bgg_id": pl.Int64,
        "bgg_user_name": pl.String,
        "bgg_user_rating": pl.Float64,
        "bgg_user_owned": pl.Boolean,
        "bgg_user_prev_owned": pl.Boolean,
        "bgg_user_for_trade": pl.Boolean,
        "bgg_user_want_in_trade": pl.Boolean,
        "bgg_user_want_to_play": pl.Boolean,
        "bgg_user_want_to_buy": pl.Boolean,
        "bgg_user_preordered": pl.Boolean,
        "bgg_user_wishlist": pl.Int64,
        "bgg_user_play_count": pl.Int64,
        "comment": pl.String,
        "published_at": pl.String,
        "updated_at": pl.String,
        "scraped_at": pl.String,
    },
)

ITEM_TYPE_SCHEMA = {
    "GameItem": GAME_ITEM_SCHEMA,
    "UserItem": USER_ITEM_SCHEMA,
    "RatingItem": RATING_ITEM_SCHEMA,
}



import argparse
import logging
import sys

from board_game_merger.config import MergeConfig
from board_game_merger.merge import merge_files

LOGGER = logging.getLogger(__name__)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Merge board game data files")

    parser.add_argument(
        "site",
        help="Site to merge data from",
    )
    parser.add_argument(
        "--item-type",
        "-t",
        choices=("GameItem", "RatingItem", "UserItem"),
        default="GameItem",
        help="Type of item to merge",
    )
    parser.add_argument(
        "--in-paths",
        "-i",
        nargs="+",
        help="Paths to input files or directories",
    )
    parser.add_argument(
        "--out-path",
        "-o",
        help="Path to output file",
    )
    parser.add_argument(
        "--clean-results",
        "-c",
        action="store_true",
        help="Clean results (remove empty fields, sort keys "
        "alphabetically and sort rows)",
    )
    parser.add_argument(
        "--latest-min-days",
        "-m",
        type=float,
        help="Minimum number of days for the latest column",
    )
    parser.add_argument(
        "--overwrite",
        "-W",
        action="store_true",
        help="Overwrite output file if it exists",
    )
    parser.add_argument(
        "--progress-bar",
        "-p",
        action="store_true",
        help="Show progress bar",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Enable verbose logging",
    )

    return parser.parse_args()


def main() -> None:
    args = _parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
        stream=sys.stdout,
    )

    merge_configs = (
        tuple(
            MergeConfig.all_sites_config(
                clean_results=args.clean_results, latest_min_days=args.latest_min_days
            )
        )
        if args.site == "all"
        else (
            MergeConfig.site_config(
                site=args.site,
                item=args.item_type,
                in_paths=args.in_paths,
                out_path=args.out_path,
                clean_results=args.clean_results,
                latest_min_days=args.latest_min_days,
            ),
        )
    )

    for merge_config in merge_configs:
        merge_files(
            merge_config=merge_config,
            overwrite=args.overwrite,
            drop_empty=True,
            sort_keys=bool(args.clean_results),
            progress_bar=bool(args.progress_bar),
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

import dataclasses
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import polars as pl

from board_game_merger.schemas import ITEM_TYPE_SCHEMA

if TYPE_CHECKING:
    from collections.abc import Generator

    from polars._typing import IntoExpr

PROJECT_DIR = Path(__file__).resolve().parent.parent.parent
FEEDS_DIR = PROJECT_DIR.parent / "board-game-scraper" / "feeds"
DATA_DIR = PROJECT_DIR.parent / "board-game-data"


@dataclasses.dataclass(frozen=True, kw_only=True)
class MergeConfig:
    schema: pl.Schema

    in_paths: str | Path | list[str] | list[Path]
    out_path: str | Path

    key_col: IntoExpr | list[IntoExpr]
    latest_col: IntoExpr | list[IntoExpr]
    latest_min: Any = None

    sort_fields: IntoExpr | list[IntoExpr] | None = None
    sort_descending: bool = False

    fieldnames_include: IntoExpr | list[IntoExpr] | None = None
    fieldnames_exclude: str | list[str] | None = None

    @classmethod
    def with_defaults(  # noqa: PLR0913
        cls,
        site: str,
        *,
        item: Literal["GameItem", "UserItem", "RatingItem"] = "GameItem",
        schema: pl.Schema | None = None,
        in_paths: str | Path | list[str] | list[Path] | None = None,
        out_path: str | Path | None = None,
        key_col: IntoExpr | list[IntoExpr] | None = None,
        latest_col: IntoExpr | list[IntoExpr] | None = None,
        clean_results: bool = False,
        latest_min_days: float | None = None,
        **kwargs: Any,
    ) -> MergeConfig:
        now = datetime.now(timezone.utc)
        now_str = now.strftime("%Y-%m-%dT%H-%M-%S")

        kwargs["schema"] = schema if schema is not None else ITEM_TYPE_SCHEMA.get(item)

        if not kwargs["schema"]:
            raise ValueError(f"Unknown item type: {item}")

        kwargs["in_paths"] = in_paths or (FEEDS_DIR / site / item)

        kwargs["key_col"] = key_col if key_col is not None else f"{site}_id"
        kwargs["latest_col"] = (
            latest_col
            if latest_col is not None
            else pl.col("scraped_at").str.to_datetime(time_zone="UTC")
        )
        if latest_min_days and latest_min_days > 0:
            kwargs.setdefault("latest_min", now - timedelta(days=latest_min_days))

        if clean_results:
            kwargs["out_path"] = out_path or (
                DATA_DIR / "scraped" / f"{site}_{item}.jl"
            )
            kwargs.setdefault("sort_fields", kwargs["key_col"])
            kwargs.setdefault(
                "fieldnames_exclude",
                ["published_at", "updated_at", "scraped_at"],
            )

        else:
            kwargs["out_path"] = out_path or (
                FEEDS_DIR / site / item / f"{now_str}-merged.jl"
            )

        return cls(**kwargs)

    @classmethod
    def site_config(
        cls,
        site: str,
        *,
        item: Literal["GameItem", "UserItem", "RatingItem"] = "GameItem",
        **kwargs: Any,
    ) -> MergeConfig:
        if site == "bgg":
            return cls.bgg_config(item=item, **kwargs)

        if item != "GameItem":
            raise ValueError(f"Unknown item type for site <{site}>: {item}")

        if site == "bgg_hotness":
            return cls.bgg_hotness_config(**kwargs)

        return cls.with_defaults(site=site, item="GameItem", **kwargs)

    @classmethod
    def bgg_config(
        cls,
        *,
        item: Literal["GameItem", "UserItem", "RatingItem"] = "GameItem",
        clean_results: bool = False,
        **kwargs: Any,
    ) -> MergeConfig:
        if item == "GameItem":
            return cls.with_defaults(
                site="bgg",
                item="GameItem",
                clean_results=clean_results,
                **kwargs,
            )

        if item == "UserItem":
            kwargs.setdefault("key_col", pl.col("bgg_user_name").str.to_lowercase())
            if clean_results:
                kwargs.setdefault("fieldnames_exclude", ["published_at", "scraped_at"])
            return cls.with_defaults(
                site="bgg",
                item="UserItem",
                clean_results=clean_results,
                **kwargs,
            )

        if item == "RatingItem":
            kwargs.setdefault(
                "key_col",
                [pl.col("bgg_user_name").str.to_lowercase(), "bgg_id"],
            )
            if clean_results:
                kwargs.setdefault("fieldnames_exclude", ["published_at", "scraped_at"])
            return cls.with_defaults(
                site="bgg",
                item="RatingItem",
                clean_results=clean_results,
                **kwargs,
            )

        raise ValueError(f"Unknown item type: {item}")

    @classmethod
    def bgg_hotness_config(
        cls,
        *,
        clean_results: bool = False,
        **kwargs: Any,
    ) -> MergeConfig:
        kwargs.setdefault(
            "key_col",
            [pl.col("published_at").str.to_datetime(time_zone="UTC"), "bgg_id"],
        )
        kwargs.setdefault(
            "sort_fields",
            [pl.col("published_at").str.to_datetime(time_zone="UTC"), "rank"],
        )
        kwargs.setdefault("fieldnames_exclude", None)

        if clean_results:
            kwargs.setdefault(
                "fieldnames_include",
                [
                    "published_at",
                    "rank",
                    "add_rank",
                    "bgg_id",
                    "name",
                    "year",
                    "image_url",
                ],
            )

        return cls.with_defaults(
            site="bgg_hotness",
            item="GameItem",
            clean_results=clean_results,
            **kwargs,
        )

    @classmethod
    def all_sites_config(
        cls,
        *,
        sites: list[str] | None = None,
        **kwargs: Any,
    ) -> Generator[MergeConfig, None, None]:
        sites = sites or [
            "bgg_hotness",
            "dbpedia",
            "luding",
            "spielen",
            "wikidata",
            "bgg",
        ]
        for site in sites:
            if site == "bgg":
                yield cls.site_config(site=site, item="GameItem", **kwargs)
                yield cls.site_config(site=site, item="UserItem", **kwargs)
                yield cls.site_config(site=site, item="RatingItem", **kwargs)
            else:
                yield cls.site_config(site=site, item="GameItem", **kwargs)
